/*
 ATtiny25 Tiny Temperature Logger - Reader (A25TTL Reader)
 Version: 1.1
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 12/11/2012
 Last Modified: 17/09/2014
 
 */

#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#ifndef boolean
	typedef uint8_t boolean;
#endif
#ifndef bool
	typedef uint8_t bool;
#endif 
#ifndef byte
	typedef uint8_t byte;
#endif

#ifndef NULL
#define NULL ((void *)0)
#endif

#define LOW 0
#define HIGH 1
#define false 0
#define true 1
#define SKIPLEDOFFDELAY 1

#define T16MS 0
#define T32MS 1
#define T64MS 2
#define T128MS 3
#define T250MS 4
#define T500MS 5
#define T1S 6
#define T2S 7
#define T4S 8
#define T8S 9

#define NOREPEATSTART 0
#define REPEATSTART 1

#define I2C_DELAY_USEC 5
#define I2C_READ 1
#define I2C_WRITE 0

// Initialize SCL/SDA pin and set the bus low
void scl_sda_low(void) {
	DDRB |= (1<<TWI_SCL_PIN);
	PORTB &= ~(1<<TWI_SCL_PIN);
	DDRB |= (1<<TWI_SDA_PIN);
	PORTB &= ~(1<<TWI_SDA_PIN);
}

// Initialize SCL/SDA pin and set the bus high
void SoftI2cMasterInit(void) {
	DDRB |= (1<<TWI_SDA_PIN);
	PORTB |= (1<<TWI_SDA_PIN);
	DDRB |= (1<<TWI_SCL_PIN);
	PORTB |= (1<<TWI_SCL_PIN);
}

// Read a byte from I2C and send Ack if more reads follow else Nak to terminate read
uint8_t SoftI2cMasterRead(uint8_t last) {
	uint8_t b = 0;
	// Make sure pull-up enabled
	PORTB |= (1<<TWI_SDA_PIN);
	DDRB &= ~(1<<TWI_SDA_PIN);
	// Read byte
	for (uint8_t i = 0; i < 8; i++) {
		// Don't change this loop unless you verify the change with a scope
		b <<= 1;
		_delay_us(I2C_DELAY_USEC);
		PORTB |= (1<<TWI_SCL_PIN);
		_delay_us(I2C_DELAY_USEC); // Extra delay (needed when 16.5MHz)
		if (bit_is_set(PINB, TWI_SDA_PIN)) b |= 1;
		PORTB &= ~(1<<TWI_SCL_PIN);
	}
	// Send Ack or Nak
	DDRB |= (1<<TWI_SDA_PIN);
	if (last) { 
		PORTB |= (1<<TWI_SDA_PIN); 
	}
	else { 
		PORTB &= ~(1<<TWI_SDA_PIN);
	}
	_delay_us(I2C_DELAY_USEC); // Extra delay (needed when 16.5MHz)
	PORTB |= (1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTB &= ~(1<<TWI_SCL_PIN);
	PORTB &= ~(1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC); // Extra delay (needed when 16.5MHz)
	return b;
}

// Write a byte to I2C
bool SoftI2cMasterWrite(uint8_t data) {
	// Write byte
	for (uint8_t m = 0x80; m != 0; m >>= 1) {
		// Don't change this loop unless you verify the change with a scope
		if (m & data) { 
			PORTB |= (1<<TWI_SDA_PIN); 
		}
		else { 
			PORTB &= ~(1<<TWI_SDA_PIN); 
		}
		_delay_us(I2C_DELAY_USEC); // Extra delay (needed when 16.5MHz)
		PORTB |= (1<<TWI_SCL_PIN);
		_delay_us(I2C_DELAY_USEC);
		PORTB &= ~(1<<TWI_SCL_PIN);
	}
	_delay_us(I2C_DELAY_USEC); // Extra delay (needed when 16.5MHz)
	// get Ack or Nak
	DDRB &= ~(1<<TWI_SDA_PIN);
	_delay_us(1); // Extra delay (needed when 16.5MHz)
	// Enable pullup
	PORTB |= (1<<TWI_SDA_PIN);
	PORTB |= (1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC); // Extra delay (needed when 16.5MHz)
	uint8_t rtn = bit_is_set(PINB, TWI_SDA_PIN);
	PORTB &= ~(1<<TWI_SCL_PIN);
	PORTB &= ~(1<<TWI_SDA_PIN);
	DDRB |= (1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC); // Extra delay (needed when 16.5MHz)
	return rtn == 0;
}

// Issue a start condition
bool SoftI2cMasterStart(uint8_t addressRW) {
	PORTB &= ~(1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTB &= ~(1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC); // Extra delay (needed when 16.5MHz)
	return SoftI2cMasterWrite(addressRW);
}

// Issue a restart condition
bool SoftI2cMasterRestart(uint8_t addressRW) {
	PORTB |= (1<<TWI_SDA_PIN);
	PORTB |= (1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC);
	return SoftI2cMasterStart(addressRW);
}

// Issue a stop condition
void SoftI2cMasterStop(void) {
	PORTB &= ~(1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTB |= (1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTB |= (1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC);
}

// Issue the device start and the address to write
bool SoftI2cMasterAddress(uint8_t deviceAddr, uint16_t addressRW, bool repeatedStart) {
	if (!SoftI2cMasterStart((deviceAddr<<1) | I2C_WRITE)) return false;
	if (!SoftI2cMasterWrite((addressRW >> 8))) return false; // MSB
	if (!SoftI2cMasterWrite((addressRW & 0xFF))) return false; // LSB
	if (repeatedStart) { if (!SoftI2cMasterRestart((deviceAddr<<1) | I2C_READ)) return false; }
	
	return true;
}

// Read 1 byte from the EEPROM device and return it
uint8_t soft_i2c_eeprom_read_byte(uint8_t deviceAddr, uint16_t readAddress) {
	uint8_t byteRead = 0;
	
	// Issue start condition, send device address, issue repeated start
	SoftI2cMasterAddress(deviceAddr, readAddress, REPEATSTART);
	
	// Read the byte
	byteRead = SoftI2cMasterRead(1);
	
	// Issue a stop condition
	SoftI2cMasterStop();
	
	return byteRead;
}

// Write 1 byte to the EEPROM
bool soft_i2c_eeprom_write_byte(uint8_t deviceAddr, uint16_t writeAddress, byte writeByte) {
	
	// Issue start condition and send device address
	SoftI2cMasterAddress(deviceAddr, writeAddress, NOREPEATSTART);
	
	// Write the byte
	if (!SoftI2cMasterWrite(writeByte)) return false;
	
	// Issue a stop condition
	SoftI2cMasterStop();
	
	// Delay 6ms for the write to complete, depends on the EEPROM you use
	_delay_ms(6);
	
	return true;
}

// Modified from http://www.arduino.cc/playground/ComponentLib/Thermistor2
// Learn more at http://www.insidegadgets.com/2011/12/18/verifying-the-thermistor-function-for-temperature-reading/
// 1 Oct 2014 - Changed to use the beta parameters of thermistors
double thermistorTemp(int RawADC) {
	double Temp = log((double) (10000 / ((10240000/(double) RawADC) - 10000)));
	
	if (Temp == 0) { // Natural log of 1 is 0 which means it's 25C, but the formula below would produce 0 as the result, so manually set 25C here
		Temp = 25;
	}
	else {
		Temp = (1 / (0.00000097670833857907 * Temp)) / ((1 / (0.00029120559114735 * Temp)) - 298.15);
		Temp = Temp - 273.15; // Convert Kelvin to Celsius
	}
	
	// Method for using coefficients
	/*double Temp;
	Temp = log((double) ((10240000/RawADC) - 10000)); // 10K thermistor with 10K resistor
	Temp = 1 / (0.001138942 + (0.000232762 + (0.000000091 * Temp * Temp )) * Temp);
	Temp = Temp - 273.15; // Convert Kelvin to Celsius
	return Temp;
	*/
	
	return Temp;
}

// Used from http://interface.khm.de/index.php/lab/experiments/sleep_watchdog_battery/
void system_sleep(void) {
	set_sleep_mode(SLEEP_MODE_PWR_DOWN); // Set sleep mode
	sleep_mode(); // System sleeps here
}

// Watchdog timeout values
// 0=16ms, 1=32ms, 2=64ms, 3=128ms, 4=250ms, 5=500ms, 6=1sec, 7=2sec, 8=4sec, 9=8sec
// Used from http://interface.khm.de/index.php/lab/experiments/sleep_watchdog_battery/
void setup_watchdog(byte ii) {
	byte bb;
	if (ii > 9 ) ii=9;
	bb=ii & 7;
	if (ii > 7) bb|= (1<<5);
	bb|= (1<<WDCE);
	
	MCUSR &= ~(1<<WDRF);
	// Start timed sequence
	WDTCR |= (1<<WDCE) | (1<<WDE);
	// Set new watchdog timeout value
	WDTCR = bb;
	WDTCR |= _BV(WDIE);
}

// Turn off the Watchdog
void turnOffwatchdog(void) {
	cli(); // Turn interrupts off
	MCUSR &= ~(1<<WDRF);
	WDTCR |= (1<<WDCE) | (1<<WDE);
	WDTCR = 0;
	sei(); // Turn interrupts back on
}

// Used and modified from Arduino wiring.c - to setup the ATtiny
void setup(void) {
	DDRB &= ~(1<<usbminusPin); // Set Inputs
	
	// Pull SCL/SDA low until a command is sent to us over USB so that the AT25TTL Logger on the other board knows not to start
	scl_sda_low(); 
	
	// Increase clock speed to 16MHz
	CLKPR = (1<<CLKPCE); // Prescaler enable
	CLKPR = 0; // Clock division factor 1 (0000)
	
	// Initialise the timer
	sbi(TIMSK, TOIE1);
	TCCR1 |= (1<<CS13) | (1<<CS12) | (1<<CS11) | (1<<CS10);
	
	sei(); // Turn on interrupts
}

// Begin V-USB default required code
/* ------------------------------------------------------------------------- */
/* ------------------------ Oscillator Calibration ------------------------- */
/* ------------------------------------------------------------------------- */

/* Calibrate the RC oscillator to 8.25 MHz. The core clock of 16.5 MHz is
 * derived from the 66 MHz peripheral clock by dividing. Our timing reference
 * is the Start Of Frame signal (a single SE0 bit) available immediately after
 * a USB RESET. We first do a binary search for the OSCCAL value and then
 * optimize this value with a neighboorhod search.
 * This algorithm may also be used to calibrate the RC oscillator directly to
 * 12 MHz (no PLL involved, can therefore be used on almost ALL AVRs), but this
 * is wide outside the spec for the OSCCAL value and the required precision for
 * the 12 MHz clock! Use the RC oscillator calibrated to 12 MHz for
 * experimental purposes only!
 */
static void calibrateOscillator(void)
{
uchar       step = 128;
uchar       trialValue = 0, optimumValue;
int         x, optimumDev, targetValue = (unsigned)(1499 * (double)F_CPU / 10.5e6 + 0.5);

    /* do a binary search: */
    do{
        OSCCAL = trialValue + step;
        x = usbMeasureFrameLength();    /* proportional to current real frequency */
        if(x < targetValue)             /* frequency still too low */
            trialValue += step;
        step >>= 1;
    }while(step > 0);
    /* We have a precision of +/- 1 for optimum OSCCAL here */
    /* now do a neighborhood search for optimum value */
    optimumValue = trialValue;
    optimumDev = x; /* this is certainly far away from optimum */
    for(OSCCAL = trialValue - 1; OSCCAL <= trialValue + 1; OSCCAL++){
        x = usbMeasureFrameLength() - targetValue;
        if(x < 0)
            x = -x;
        if(x < optimumDev){
            optimumDev = x;
            optimumValue = OSCCAL;
        }
    }
    OSCCAL = optimumValue;
}
/*
Note: This calibration algorithm may try OSCCAL values of up to 192 even if
the optimum value is far below 192. It may therefore exceed the allowed clock
frequency of the CPU in low voltage designs!
You may replace this search algorithm with any other algorithm you like if
you have additional constraints such as a maximum CPU clock.
For version 5.x RC oscillators (those with a split range of 2x128 steps, e.g.
ATTiny25, ATTiny45, ATTiny85), it may be useful to search for the optimum in
both regions.
*/

void hadUsbReset(void)
{
	calibrateOscillator();
}
// End V-USB default required code

// Start USB communications
void startUSB(void) {
	usbDeviceDisconnect();
	_delay_ms(300); // 300 ms disconnect
	usbDeviceConnect();
	usbInit();
}
